/*
 functions for communication between pic and host - via usb, rs232 or both

 $Id: io.c,v 1.3 2019/07/07 10:34:45 daeb Exp $
*/

/*

 Copyright (C) 2019 Bernhard Dubler ( klappe@daeublers.de )

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "mcc_generated_files/mcc.h"
#include "def.h"
#include "io.h"

#include <stdint.h>

#ifdef USEUSB
static uint8_t readBuffer[64];
static uint8_t numBytesRead=0, rindex=0;
static uint8_t writeBuffer[64];
#endif

#define MAX_DIGITS 12
static char digits[MAX_DIGITS];

#define  testchar_rs232() (EUSART_is_rx_ready())

#define is_usb_ready() ((USBGetDeviceState()>=CONFIGURED_STATE) && (USBIsDeviceSuspended()!=true))

#ifdef USEUSB
void service_io(void)
{		if (is_usb_ready()) CDCTxService();
}

static int testchar_usb(void)
{	if (numBytesRead>0)
		return 1;
	if( !is_usb_ready())
		return 0;
    numBytesRead = getsUSBUSART(readBuffer, sizeof(readBuffer));
	rindex=0;
	return  (numBytesRead>0)?1:0;
}

static char getchar_usb(void)
{	while (!is_usb_ready());
	while (!testchar_usb()) CDCTxService();
	numBytesRead--;
	return readBuffer[rindex++];
}

static void putchar_usb(char ch)
{   if (!is_usb_ready()) return;
	while ( USBUSARTIsTxTrfReady() == false) CDCTxService();
	writeBuffer[0]=ch;
	putUSBUSART(writeBuffer,1);
}

static void putstr_usb(const char *s)
{	while (*s) putchar_usb(*s++);
}

#endif // USB

#ifdef USERS232
static char getchar_rs232(void)
{	while (!EUSART_is_rx_ready());
	return EUSART_Read();
}


static void putchar_rs232(char ch)
{	while (!EUSART_is_tx_ready());
	EUSART_Write(ch);
}


static void putstr_rs232(const char *s)
{	while (*s) putchar_rs232(*s++);
}
#endif // USERS232


#if defined(USERS232) && defined(USEUSB)
int testchar(void)
{	return testchar_rs232() || testchar_usb();
}
#else
#ifdef USERS232
int testchar(void)
{	return testchar_rs232();
}
#endif 
#ifdef USEUSB
int testchar(void)
{	return testchar_usb();
}
#endif 
#endif

char getchar(void)
{	int ch=-1;
	do
	{
#ifdef USERS232
		if (testchar_rs232())
			ch=(uint8_t)getchar_rs232();
#endif
#if defined(USERS232) && defined(USEUSB)
		else
#endif
#ifdef USEUSB
		if (testchar_usb())
			ch=(uint8_t)getchar_usb();
#endif
	}while (ch<0);
	return (char)ch;
}

void putchar(char ch)
{	
#ifdef USEUSB
	if (is_usb_ready()) putchar_usb(ch);
#endif
#ifdef USERS232
	putchar_rs232(ch);
#endif
}

void putstr(const char *s)
{	
#ifdef USEUSB
	if (is_usb_ready()) putstr_usb(s);
#endif
#ifdef USERS232
	putstr_rs232(s);
#endif
}

static void d64(unsigned long *hp, unsigned long *lp, uint8_t *r)
{   unsigned long h=*hp, l=*lp;
    unsigned long q1, r1;
    unsigned long q2=0, r2;
    unsigned long q3;
	uint8_t r3;
    unsigned long t;

    r1=h%10;
    q1=h/10;
  //printf("   q1=%lx r1=%lx\r\n", q1, r1);
    if (r1)
    {   t=(r1<<4) + (l>>28);
      //printf("   t=%lx\r\n", t);
        q2=t/10;
        r2=t%10;
      //printf("   q2=%lx r2=%lx\r\n", q2, r2);
        l= (l&(((unsigned long)-1)>>(32-28))) | (r2<<28);
      //printf("   l=%lx\r\n", l);
    }
    q3=l/10;
    r3=l%10;
  //printf("   q3=%lx r3=%x\r\n", q3, (unsigned)r3);
    *hp=q1;
    *lp=(q2<<28) + q3;
    *r=r3;
}

void print64(unsigned long h, unsigned long l, int8_t mindigits)
{	uint8_t r;
	int8_t j,i;

	j=0;
	if (!h && !l) putchar('0'), mindigits--;
	while (h || l || mindigits>0)
	{	d64(&h, &l, &r);
		digits[j++]=r+'0';
		mindigits--;
	}
	for (i=j-1; i>=0; i--) putchar(digits[i]);
}


void print16(unsigned l, int8_t mindigits)
{	print64(0, l, mindigits);
}


#ifdef DEBUGRFID
void print32(unsigned long l, int8_t mindigits)
{	print64(0, l, mindigits);
}

/*
void print32(unsigned long l, int8_t mindigits)
{	uint8_t r;
	int8_t j,i;

	j=0;
	if (!l) putchar('0'), mindigits--;
	while (l || mindigits>0)
	{	r=l%10;
		l/=10;
		digits[j++]=r+'0';
		mindigits--;
	}
	for (i=j-1; i>=0; i--) putchar(digits[i]);
}
*/

void printhex(unsigned long l, int8_t mindigits)
{
	unsigned char ch;
	int8_t j,i;

	j=0;
	if (!l) putchar('0'), mindigits--;
	while (l || mindigits>0)
	{	ch=(l&0xf)+'0';
		if (ch>'9') ch+='a'-'9'-1;
		digits[j++]=ch;
		l>>=4;
		mindigits--;
	}
	for (i=j-1; i>=0; i--) putchar(digits[i]);
}
#endif
