/*
 rfid functions

 $Id: rfid.c,v 1.5 2019/07/07 10:34:45 daeb Exp $
*/

/*

 Copyright (C) 2019 Bernhard Dubler ( klappe@daeublers.de )

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "mcc_generated_files/mcc.h"
#include "io.h"
#include "def.h"
#include "rfid.h"
#include "time.h"

#define TIMER_FREQ (_XTAL_FREQ/32/4)

#define BIT_FREQ (134000/32)		/* 134 kHz and 32 cycles per bit */
#define PLUSE_LONG (TIMER_FREQ/BIT_FREQ)	/* nominal length of a long pulse - a one-bit */
#define PLUSE_SHORT (PLUSE_LONG/2)			/* nominal length of a short pulse - half of a zero-bit */

#define PLUSE_TOSHORT (PLUSE_SHORT/3.7)		/* upper limit for a long pulse */
#define PLUSE_TOLONG (PLUSE_LONG*1.3)		/* lower limit for a short pulse */
#define PLUSE_MINLONG ((PLUSE_LONG+PLUSE_SHORT)/2) /* upper limit for short and lower limit for long pulse */

#define RFID_DELTA_TICKS (1*TICKS_PER_S)	/* min time between 2 reported rfid reads = 1s */
#define INFRAME_LED_TICKS (8*TICKS_PER_S)	/* how long the inframe led will be turned on = 8s */

#ifdef DEBUGRFID
unsigned int Sflag=0;
int nirq_SenOut=0;
static int nirq_SenOut_last=0;
static int nirq_high=0;
static int nirq_low=0;
static int nirq_low_last=0;
static int nirq_nochange=0;
static int nirq=0, nbits=0;
static int rbit=0, wbit=0;
#define MAX_BITS 400
static char bits[MAX_BITS];
#define DEBUGEX(x) x
#else
#define DEBUGEX(x) 
#endif

static uint8_t LastSigState;			/* last state of the comperator **/
#define MAX_BYTES 10
static uint8_t bytes[MAX_BYTES];		/* to store bits during read */
static uint8_t erg[MAX_BYTES];			/* result is copied there after read is complete */
static uint8_t rfid_ready=0;			/* flag that valid rfid is stored in erg */
static timetick_t last_rfid_time, inframe_time;

static uint8_t lasts=0;					/* last pulse was a short pulse */
static uint8_t state=0;					/* counts the read bits from 0 to STATE_ENOUGH */
#define STATE_PRECOMPLETE 11			/* bits of preamble */
#define STATE_ENOUGH 11+9*10			/* preamble + 10 bytes x (8 bits + 1 bit as frame ) */
static uint8_t bc;						/* byte counter */

static void rfid_irq(void)
{	
	uint8_t td1;
	uint8_t SigState;
#ifdef DEBUGRFID
	uint8_t dbit;	/* either s, 1, Y (= 1 after s), X (too long), x (too short) */
#endif
	uint8_t bit;	/* either 0, 1, Y (= 1 after s), y (half 0 ), X (too long), x (too short) */
	static uint8_t bn, nzero=0;

	if ((SigState=CMP2_GetOutputStatus())!=LastSigState && state<STATE_ENOUGH)
	{	td1=TMR0_ReadTimer();
		TMR0_WriteTimer(0);
		{	if (td1>(uint8_t)PLUSE_TOLONG)
				DEBUGEX(dbit=)bit='X',lasts=0;
			else if (td1>(uint8_t)PLUSE_MINLONG)
			{	if (lasts)
					DEBUGEX(dbit=)bit='Y',lasts=0;
				else
					DEBUGEX(dbit=)bit='1';
			}
			else if (td1>(uint8_t)PLUSE_TOSHORT)
			{	if (lasts)
				{	lasts=0, bit='0';
#ifdef DEBUGRFID
					dbit='s';
#endif
				}
				else
				{	lasts=1, bit='y';
#ifdef DEBUGRFID
					dbit='s';
#endif
				}
			}
			else
				DEBUGEX(dbit=)bit='x',lasts=0;
#ifdef DEBUGRFID
			bits[wbit++]=dbit;
			if (wbit==MAX_BITS) wbit=0;
			if (nbits<MAX_BITS) nbits++;
			else rbit=wbit;
#endif
		}
		if (bit=='y')	/* read half bit */
		{	if (state==STATE_PRECOMPLETE-1)
				lasts=0; 	/* after 20 's' every 's' is wellcome */
		}
		else			/* got full bit - or error */
		{	
			if (bit=='0') nzero++;	/* count the 0-bit */
			else nzero=0;			/* if we get a frame error, 0-bit may be part of new peramble */
			if (bit=='0' && state<=STATE_PRECOMPLETE-1)
			{	if (state<STATE_PRECOMPLETE-1) state++;
			}
			else if (bit=='1' && state==STATE_PRECOMPLETE-1) /* preamble complete */
			{	for (bc=0; bc<MAX_BYTES; bc++) bytes[bc]=0;
				state++,bn=bc=0;
			}
			else if ((bit=='0' || bit=='1') && state>=STATE_PRECOMPLETE && state<STATE_ENOUGH)
			{	if (bn<8)
				{	if (bit=='1') bytes[bc]|=(1<<bn);
					bn++, state++;
				}
				else if (bit!='1')	/* after 8 bits a 1-bit is expected */
					state=nzero;	/* framing error - from beginning */
				else
					bn=0, bc++, state++;
			}
			else if (state<STATE_ENOUGH)	/* error - from beginning */
				state=0;
		}
#ifdef DEBUGRFID
		nirq++;
		if (SigState)
		{	nirq_high++;
		}
		else
		{	nirq_low++;
		}
#endif
		if (state==STATE_ENOUGH)
		{	uint8_t i;

			LED_InFrame_SetLow();
			inframe_time=now_timeticks;
			for (i=0; i<MAX_BYTES && erg[i]==bytes[i]; i++) ;
			if (i<MAX_BYTES || now_timeticks-last_rfid_time>RFID_DELTA_TICKS)
			{	for (i=0; i<MAX_BYTES; i++)
					erg[i]=bytes[i];
				last_rfid_time=now_timeticks;
				rfid_ready=1;
			}
#ifndef DEBUGRFID
			lasts=0;
			bc=0;
			state=0;
#endif
		}
		LastSigState=SigState;
		//if (state>=STATE_PRECOMPLETE && state<STATE_ENOUGH)  LED_InFrame_SetLow();
		//else  LED_InFrame_SetHigh();
	}
#ifdef DEBUGRFID
	else
	{	nirq_nochange++;
		/*
		bits[wbit++]=SigState?'P':'p';
		if (wbit==MAX_BITS) wbit=0;
		if (nbits<MAX_BITS) nbits++;
		else rbit=wbit;
		*/
	}
#endif
}

/// Get the lower 8 bits macro
#define lo8(x) ((x)&0xff)
/// Get the higher 8 bits macro
#define hi8(x) ((x)>>8) 

/**
 * This function calculates the CCITT crc (16-bit) of a byte
 * @param crc The previous crc value
 * @param data The data to calculate the crc over
 * @return The new crc value
 */
unsigned short crc_ccitt_update(unsigned short crc, unsigned char data)
{
	data ^= lo8(crc);
    data ^= data << 4;

    return ((((unsigned short)data << 8) | hi8(crc)) ^ (unsigned char)(data >> 4) 
            ^ ((unsigned short)data << 3));
}


void print_erg(void)
{
	unsigned short crc=0;
	unsigned long l,h;
	unsigned r;
	uint8_t i;

	for (i=0;i<bc;i++)
		crc=crc_ccitt_update(crc, erg[i]);
	r=((unsigned)erg[5]<<2) | ((unsigned)erg[4]>>6);
	putstr("id: ");
	print16(r,3);
	putchar('.');
	h=erg[4]&0x3f;
	l=(((((unsigned long)erg[3]<<8) | (unsigned long)erg[2])<<8 | (unsigned long)erg[1])<<8) | (unsigned long)erg[0];
	print64(h,l,12);
#ifdef DEBUGRFID
	putstr(" crc: ");
	if (crc) print16(crc,0);
	else putstr("ok");
#else
	if (crc) putstr(" bad crc");
#endif
	putstr("\r\n");
}


#ifdef DEBUGRFID
void print_rfid_state(void)
{	putstr("nirq_high="); print32(nirq_high,0);
	putstr(" nirq_low="); print32(nirq_low,0);
	putstr(" nirq_nochange="); print32(nirq_nochange,0);
	putstr(" nirq_SenOut="); print32(nirq_SenOut,0);
	putstr(" state="); print16(state,0);
	putstr(" nbits="); print16(nbits,0);
	putstr(" bc="); print16(bc,0);
	putstr("\r\n");
}

void print_bitbuffer(void)
{	int i;
			
	putstr("rbit="); print16(rbit,0);
	putstr(" wbit="); print16(wbit,0);
	putstr("\r\n");
	while (nbits)
	{	i=rbit+1;
		if (i==MAX_BITS) i=0;
		if (bits[rbit]=='s' && nbits>1 && bits[i]=='s')
		{	putchar('0');
			i++;
			if (i==MAX_BITS) i=0;
			nbits--;
		}
		else
		{	putchar(bits[rbit]);
		}
		rbit=i;
		nbits--;
	}
	putstr("\r\n");
}
#endif


void service_rfid(void)
{	static uint8_t first_run=1;

	if (first_run)
	{
#ifdef DEBUGRFID
		putstr("\r\nstart SFlag="); print16(Sflag,0); putstr("\r\n");
		putstr("PULSE_SHORT="); printhex(PLUSE_SHORT,0); putstr("\r\n");
		putstr("PULSE_MINLONG="); printhex(PLUSE_MINLONG,0); putstr("\r\n");
		putstr("PULSE_TOSHORT="); printhex(PLUSE_TOSHORT,0); putstr("\r\n");
		putstr("PULSE_TOLONG="); printhex(PLUSE_TOLONG,0); putstr("\r\n");
#else
		putstr("\r\nstart\r\n");
#endif
		first_run=0;
	}

	if (now_timeticks-inframe_time>INFRAME_LED_TICKS) LED_InFrame_SetHigh();
	//if (state>=STATE_PRECOMPLETE && state<STATE_ENOUGH)  LED_InFrame_SetLow();
	//else  LED_InFrame_SetHigh();

#ifdef DEBUGRFID
	if (state==STATE_ENOUGH)
	{	uint8_t i;

		putstr("\r\nstop\r\n");
		print_rfid_state();
		print_bitbuffer();
		putstr("\r\n");
		for (i=0; i<bc; i++)
		{		print16(bytes[i], 0); putstr("\r\n");
		}
		nirq_high=nirq_low=nirq_nochange=nirq=nbits=0; nirq_low_last=0;
		nirq_SenOut=nirq_SenOut_last=0;
		rbit=0, wbit=0;
		lasts=0;
		state=0;
		bc=0;

		LastSigState=CMP2_GetOutputStatus();
	}
#endif
	if (rfid_ready)
	{	print_erg();
		rfid_ready=0;
	}
#ifdef DEBUGRFID
	if (Sflag && (nirq_low!=nirq_low_last || nirq_SenOut!=nirq_SenOut_last))
	{	print_rfid_state();
		if (nirq_SenOut!=nirq_SenOut_last)
		{	print_bitbuffer();
			rbit=0, wbit=0;
			nbits=0;
			nirq_low_last=nirq_low=0;
			nirq_nochange=0;
		}
		nirq_low_last=nirq_low;
		nirq_SenOut_last=nirq_SenOut;
	}
#endif
}

void init_rfid()
{	CMP2_SetInterruptHandler(rfid_irq);

	LED_InFrame_SetHigh();

	LastSigState=CMP2_GetOutputStatus();
#ifdef DEBUGRFID
	nirq_high=nirq_low=nirq=nbits=0; nirq_low_last=0;
	rbit=0, wbit=0;
#endif
	lasts=0;
	state=0;
	bc=0;
	TMR0_WriteTimer(0);
}
