/*
 handles the two reed-switches and the motion sensor

 $Id: sensors.c,v 1.6 2019/07/07 10:34:45 daeb Exp $
*/
/*

 Copyright (C) 2019 Bernhard Dubler ( klappe@daeublers.de )

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "mcc_generated_files/mcc.h"
#include "def.h"
#include "sensors.h"
#include "time.h"
#include "io.h"
#ifdef DEBUGRFID
#include "rfid.h"
#endif

#define MAX_EVENTS 30

#define IDLE_SECS 10	/* seconds without event to report event */
#define IDLE_TICKS (IDLE_SECS*TICKS_PER_S)
#define OUT_DEDECT_MS	1000	/* t(CLOSED_ON)>OUT_DEDECT_MS && t(OPEN_ON)-t(CLOSED_ON)<OUT_DEDECT_MS */
#define OUT_DEDECT_TICKS	((unsigned long)OUT_DEDECT_MS*TICKS_PER_S/1000)

#define CLOSED_ON 0
#define CLOSED_OFF 1
#define OPEN_ON 2
#define OPEN_OFF 3

#define S_CLOSED 0
#define S_OPEN_OUT 1
#define S_OPEN_IN 2

#ifdef DEBUGSENSORS
static char *event_text[]={"closed-on", "closed-off", "open-on", "open-off"};
#endif

static struct
	{	uint8_t type;
		timetick_t time;
	} event[MAX_EVENTS];
static uint8_t events=0;
static uint8_t state;

#define OPENED_TRIGGER_DELTA_TICKS (5*TICKS_PER_S)
#define MOTION_TRIGGER_DELTA_TICKS (2*TICKS_PER_S)

static uint8_t motion_trigger=0;
static uint8_t opened_trigger=0;
static timetick_t last_motion_time;
static timetick_t last_opened_time;
static uint8_t last_opened_type, last_closed_type, last_motion_state;

/*
static void SenMotion_irq(void)
{	
	if (!motion_trigger && now_timeticks-last_motion_time>MOTION_TRIGGER_DELTA_TICKS)
	{	motion_trigger=1;
		last_motion_time=now_timeticks;
	}
}
*/

static void SenClosed_irq(void)
{	uint8_t type;

	type=SenClosed_GetValue()?CLOSED_OFF:CLOSED_ON;

	if (type!=last_closed_type)
	{	if (type==CLOSED_OFF && !opened_trigger && now_timeticks-last_opened_time>OPENED_TRIGGER_DELTA_TICKS)
		{	opened_trigger=1;
			last_opened_time=now_timeticks;
		}
		if (events<MAX_EVENTS)
		{	event[events].type=type;
			event[events++].time=now_timeticks;
		}
		last_closed_type=type;
#ifdef DEBUGSENSORS
		putstr("closed_trigger ");
		putstr(event_text[type]);
		putstr("\r\n");
#endif
	}
}

static void SenOpened_irq(void)
{	uint8_t type;

#ifdef DEBUGRFID
	nirq_SenOut++;
#else
	type=SenOpened_GetValue()?OPEN_OFF:OPEN_ON;

	if (type!=last_opened_type)
	{	if (events<MAX_EVENTS)
		{	event[events].type=type;
			event[events++].time=now_timeticks;
		}
		last_opened_type=type;
#ifdef DEBUGSENSORS
		putstr("open_trigger ");
		putstr(event_text[type]);
		putstr("\r\n");
#endif
	}
#endif
}

void init_sensors(void)
{   /* poll the lines periodically instead of irq on change */
	/* irqs are not working well */

	//IOCAF4_SetInterruptHandler(SenMotion_irq);
    //IOCBF4_SetInterruptHandler(SenClosed_irq);
    //IOCBF6_SetInterruptHandler(SenOpened_irq);

	state=SenClosed_GetValue()?S_OPEN_IN:S_CLOSED;
	last_opened_type=SenOpened_GetValue()?OPEN_OFF:OPEN_ON;
	last_closed_type=SenClosed_GetValue()?CLOSED_OFF:CLOSED_ON;
	last_motion_state=SenMotion_GetValue();
}

void service_sensors(void)
{
	int8_t i, o, c, c2, out;
	timetick_t log_time, now;
	char *log_text;

#ifdef DEBUGSENSORS
	static uint8_t firstrun=1;

	if (firstrun)
	{	putstr("startup - closed=");
		print16(SenClosed_GetValue(), 0);
		putstr(" opened=");
		print16(SenOpened_GetValue(), 0);
		putstr("\r\n");
		firstrun=0;
	}
#endif

	SenClosed_irq();
	SenOpened_irq();
	if (events>0)							/* some reed-switch events stored ? */
	{	now=now_timeticks;   
		if (now-event[events-1].time>IDLE_TICKS) /* idle for some time - then find out what happened */
		{
#ifdef DEBUGSENSORS
			putstr("klappe idle\r\n");
			putstr("-----------\r\n");
			for (i=0; i<events; i++)
			{	print16(i,0);
				putstr(" msecs=");
				print16((event[i].time-event[0].time)*(1000/TICKS_PER_S),0);
				putchar(' ');
				putstr(event_text[event[i].type]);
				putstr("\r\n");
			}
#endif

			out=1;
			log_text=NULL;
			if (state!=S_CLOSED && event[events-1].type==CLOSED_ON && event[events-1].time-event[0].time<OUT_DEDECT_TICKS)
			{	log_time=event[events-1].time;
				log_text=(state==S_OPEN_OUT?"out close":"in close");
				state=S_CLOSED;
			}
			else
			{	log_time=event[0].time;
				for (o=0; o<events && event[o].type!=OPEN_ON; o++);
				if (o<events)	/* open sensor was triggered */
				{
					for (c=o-1; c>=0 && event[c].type!=CLOSED_ON; c--); /* search last CLOSED_ON backwards */
					for (c2=o+1; c2<events && event[c2].type!=CLOSED_ON; c2++); /* search next CLOSED_ON forwards */
					if (c>=0 && event[c].time-event[0].time>=OUT_DEDECT_TICKS &&
							event[o].time-event[c].time<OUT_DEDECT_TICKS &&
						(c2>=events || event[c2].time-event[o].time<OUT_DEDECT_TICKS))
					{	out=0;
#ifdef DEBUGSENSORS
						putstr("->  o=");
						print16(o,0);
						putstr(" c=");
						print16(c,0);
						putstr(" in dedect\r\n");
#endif
					}
#ifdef DEBUGSENSORS
					else
					{	putstr("->  o=");
						print16(o,0);
						putstr(" c=");
						print16(c,0);
						putstr(" out dedect\r\n");
					}
#endif
				}
				else
				{	out=0;
#ifdef DEBUGSENSORS
					putstr("->  in dedect\r\n");
#endif
				}
				if (event[events-1].type==CLOSED_ON)
					log_text=(out?"out":"in");
				else
				{	state=out?S_OPEN_OUT:S_OPEN_IN;
					log_text=(out?"out open":"in open");
				}
			}
			if (log_text)
			{	
				putstr(log_text);
				putchar(' ');
				print16((now-log_time)/TICKS_PER_S, 0);
				putstr("s ago\r\n");
			}
			events=0;
		}
	}

	i=SenMotion_GetValue();
	if (i==0 && last_motion_state==1 && motion_trigger==0)
	{	motion_trigger=2;
		last_motion_time=now_timeticks;
		putstr("motion\r\n");
	}
	else if (motion_trigger==2 && now_timeticks-last_motion_time>MOTION_TRIGGER_DELTA_TICKS)
	{	motion_trigger=0;
	}
	last_motion_state=i;

	if (opened_trigger==1)
	{	opened_trigger=2;
		putstr("opened\r\n");
	}
	else if (opened_trigger==2 && now_timeticks-last_opened_time>OPENED_TRIGGER_DELTA_TICKS)
	{	opened_trigger=0;
	}
}
